import enum
from abc import ABC, abstractmethod
from typing import Final, Literal, Self

from verry import function as vrf
from verry.misc.formatspec import FormatSpec
from verry.typing import Scalar, SignedComparable


class RoundingMode(enum.Enum):
    """Rounding mode specifier.

    Attributes
    ----------
    ROUND_CEILING
    ROUND_FAST
    ROUND_FLOOR
    """

    ROUND_CEILING = enum.auto()
    ROUND_FAST = enum.auto()
    ROUND_FLOOR = enum.auto()

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}>"


ROUND_CEILING: Final = RoundingMode.ROUND_CEILING
ROUND_FAST: Final = RoundingMode.ROUND_FAST
ROUND_FLOOR: Final = RoundingMode.ROUND_FLOOR


class Converter[T: SignedComparable](ABC):
    """Provides numeric/numeric and numeric/string conversions."""

    __slots__ = ()

    @abstractmethod
    def fromfloat(self, value: float, strict: bool = True) -> T:
        """Convert the float to a number.

        Raises
        ------
        ValueError
            If `strict` is ``True`` and `value` cannot be converted without loss.
        """
        raise NotImplementedError

    @abstractmethod
    def fromstr(self, value: str, rounding: RoundingMode) -> T:
        """Convert the string to a number with rounding taken into account.

        Raises
        ------
        ValueError
            If `value` does not represent a number.
        """
        raise NotImplementedError

    def tostr(
        self,
        value: T,
        rounding: Literal[RoundingMode.ROUND_CEILING, RoundingMode.ROUND_FLOOR],
    ) -> str:
        """Convert the number to a string with rounding taken into account."""
        raise NotImplementedError

    def format(
        self,
        value: T,
        spec: FormatSpec,
        rounding: Literal[RoundingMode.ROUND_CEILING, RoundingMode.ROUND_FLOOR],
    ) -> str:
        raise NotImplementedError

    def fromint(self, value: int, rounding: RoundingMode) -> T:
        """Convert the integer to a number with rounding taken into account.

        This method is defined as ``fromstr(str(value), rounding)`` if not overloaded.
        """
        return self.fromstr(str(value), rounding)

    def repr(self, value: T) -> str:
        raise NotImplementedError


class Operator[T: SignedComparable](ABC):
    """Provides arithmetic with directed rounding and basic constants.

    Notes
    -----
    Classes that inherit from this must define the class constants `ZERO`, `ONE`, and
    `INFINITY`.
    """

    __slots__ = ()
    ZERO: T
    ONE: T
    INFINITY: T

    @abstractmethod
    def cadd(self, lhs: T, rhs: T) -> T:
        """Add and round towards positive infinity."""
        raise NotImplementedError

    def fadd(self, lhs: T, rhs: T) -> T:
        """Add and round towards negative infinity.

        This method is defined as ``-cadd(-lhs, -rhs)`` if not overloaded.
        """
        return -self.cadd(-lhs, -rhs)

    def csub(self, lhs: T, rhs: T) -> T:
        """Subtract and round towards positive infinity.

        This method is defined as ``cadd(lhs, -rhs)`` if not overloaded.
        """
        return self.cadd(lhs, -rhs)

    def fsub(self, lhs: T, rhs: T) -> T:
        """Subtract and round towards negative infinity.

        This method is defined as ``-cadd(-lhs, rhs)`` if not overloaded.
        """
        return -self.cadd(-lhs, rhs)

    @abstractmethod
    def cmul(self, lhs: T, rhs: T) -> T:
        """Multiply and round towards positive infinity."""
        raise NotImplementedError

    def fmul(self, lhs: T, rhs: T) -> T:
        """Multiply and round towards negative infinity.

        This method is defined as ``-cmul(-lhs, rhs)`` if not overloaded.
        """
        return -self.cmul(-lhs, rhs)

    @abstractmethod
    def cdiv(self, lhs: T, rhs: T) -> T:
        """Divide and round towards positive infinity."""
        raise NotImplementedError

    def fdiv(self, lhs: T, rhs: T) -> T:
        """Divide and round towards negative infinity.

        This method is defined as ``-cdiv(-lhs, rhs)`` if not overloaded.
        """
        return -self.cdiv(-lhs, rhs)

    @abstractmethod
    def csqr(self, value: T) -> T:
        """Calcurate the square root and round towards positive infinity."""
        raise NotImplementedError

    @abstractmethod
    def fsqr(self, value: T) -> T:
        """Calcurate the square root and round towards negative infinity."""
        raise NotImplementedError


class Interval[T: SignedComparable](Scalar, ABC):
    """Abstract base class for inf-sup type intervals.

    Parameters
    ----------
    inf : endtype | float | int | str | None, optional
        Infimum of the interval.
    sup : endtype | float | int | str | None, optional
        Supremum of the interval.

    Attributes
    ----------
    inf : endtype
        Infimum of the interval.
    sup : endtype
        Supremum of the interval.
    converter : Converter
    endtype : type[endtype]
    operator : Operator

    Notes
    -----
    Classes that inherit from this must define class constants `converter`, `operator`,
    and `endtype`, where `endtype` is a type of endpoints.
    """

    __slots__ = ("inf", "sup")
    inf: T
    sup: T
    converter: Converter[T]
    endtype: type[T]
    operator: Operator[T]

    def __init__(
        self,
        inf: T | float | int | str | None = None,
        sup: T | float | int | str | None = None,
    ):
        if inf is None:
            if sup is None:
                self.inf = self.sup = self.operator.ZERO
                return

            inf = sup

        match inf:
            case self.endtype():
                self.inf = inf

            case str():
                self.inf = self.converter.fromstr(inf, ROUND_FLOOR)

            case int():
                self.inf = self.converter.fromint(inf, ROUND_FLOOR)

            case float():
                self.inf = self.converter.fromfloat(inf)

            case _:
                raise TypeError

        if sup is None:
            self.sup = self.inf
            return

        match sup:
            case self.endtype():
                self.sup = sup

            case str():
                self.sup = self.converter.fromstr(sup, ROUND_CEILING)

            case int():
                self.sup = self.converter.fromint(sup, ROUND_CEILING)

            case float():
                self.sup = self.converter.fromfloat(sup)

            case _:
                raise TypeError

        if self.inf > self.sup:
            raise ValueError

    @classmethod
    def ensure(cls, value: Self | T | float | int | str) -> Self:
        """Convert `value` to an interval and return its copy."""
        return value.copy() if isinstance(value, cls) else cls(value)  # type: ignore

    def copy(self) -> Self:
        """Return a shallow copy of the interval."""
        return self.__class__(self.inf, self.sup)

    def diam(self) -> T:
        """Return an upper bound of the diameter.

        Warning
        -------
        ``x.diam()`` might not be finite even if `x` is bounded.

        See Also
        --------
        isbounded

        Examples
        --------
        >>> from verry import FloatInterval
        >>> x = FloatInterval(-1, 1)
        >>> x.diam()
        2.0
        >>> y = 1e+308 * x
        >>> y.isbounded()
        True
        >>> y.diam()  # results in a positive infinity due to overflow.
        inf
        """
        return self.operator.csub(self.sup, self.inf)

    def hull(self, *args: Self | T | float | int) -> Self:
        """Return an interval hull."""

        result = self.__class__(self.inf, self.sup)

        for arg in args:
            result |= arg

        return result

    def interiorcontains(self, other: Self | T | float | int) -> bool:
        """Return ``True`` if the interior of the interval contains `other`."""
        match other:
            case self.endtype():
                return self.inf < other < self.sup

            case self.__class__():
                return self.inf < other.inf and other.sup < self.sup

            case int():
                return self.interiorcontains(self.__class__(other))

            case float():
                return self.interiorcontains(self.converter.fromfloat(other))

        raise TypeError

    def isbounded(self) -> bool:
        """Return ``True`` if both `inf` and `sup` are finite."""
        return self.mag() < self.operator.INFINITY

    def isdisjoint(self, other: Self | T | float | int) -> bool:
        """Return ``True`` if the interval has no elements in common with `other`."""
        match other:
            case self.endtype():
                return self.inf > other or self.sup < other

            case self.__class__():
                return self.inf > other.sup or self.sup < other.inf

            case int():
                return self.isdisjoint(self.__class__(other))

            case float():
                return self.isdisjoint(self.converter.fromfloat(other))

        raise TypeError

    def issubset(self, other: Self) -> bool:
        """Test whether every element in the interval is in `other`."""
        if type(self) is not type(other):
            return False

        return self.inf >= other.inf and self.sup <= other.sup

    def issuperset(self, other: Self) -> bool:
        """Test whether every element in `other` is in the interval."""
        if type(self) is not type(other):
            return False

        return self.inf <= other.inf and self.sup >= other.sup

    def mag(self) -> T:
        """Return a magnitude of the interval.

        The magnitude of `x` is defined as ``max(abs(x.inf), abs(x.sup))``.
        """
        return max(abs(self.inf), abs(self.sup))

    @abstractmethod
    def mid(self) -> T:
        """Return an approximation of the midpoint.

        ``x.mid() in x`` is guaranteed to be ``True`` for any `x`.
        """
        raise NotImplementedError

    def mig(self) -> T:
        """Return a mignitude of the interval.

        The mignitude of `x` is defined as ``min(abs(x.inf), abs(x.sup))``.
        """
        return min(abs(self.inf), abs(self.sup))

    def rad(self) -> T:
        """Return an upper bound of the radius."""
        return (self - self.mid()).mag()

    def width(self) -> T:
        """Alias of :meth:`diam`."""
        return self.diam()

    def _verry_overload_(self, fun, *args, **kwargs):
        match fun:
            case vrf.sqrt:
                return self.__sqrt()

        return NotImplemented

    def __repr__(self) -> str:
        try:
            inf = self.converter.repr(self.inf)
            sup = self.converter.repr(self.sup)
            return f"{type(self).__name__}(inf={inf}, sup={sup})"
        except NotImplementedError:
            return super().__repr__()

    def __str__(self) -> str:
        try:
            inf = self.converter.tostr(self.inf, ROUND_FLOOR)
            sup = self.converter.tostr(self.sup, ROUND_CEILING)
            return f"[{inf}, {sup}]"
        except NotImplementedError:
            return self.__repr__()

    def __format__(self, format_spec: str) -> str:
        spec = FormatSpec(format_spec)

        try:
            tmp = spec.replace(fill="\u0020", align=None, zfill=False, width=None)
            inf = self.converter.format(self.inf, tmp.copy(), ROUND_FLOOR)
            sup = self.converter.format(self.sup, tmp, ROUND_CEILING)
        except NotImplementedError:
            raise ValueError

        spec = FormatSpec(fill=spec.fill, align=spec.align, width=spec.width)
        return spec.format(f"[{inf}, {sup}]")

    def __eq__(self, other) -> bool:
        if type(other) is not type(self):
            return NotImplemented

        return other.inf == self.inf and other.sup == self.sup

    def __contains__(self, item) -> bool:
        match item:
            case self.endtype():
                INFINITY = self.operator.INFINITY
                return abs(item) < INFINITY and self.inf <= item <= self.sup

            case int():
                return self.__class__(item).issubset(self)

            case float():
                return self.__contains__(self.converter.fromfloat(item))

        raise TypeError

    def __add__(self, rhs: Self | T | float | int) -> Self:
        cadd = self.operator.cadd
        fadd = self.operator.fadd

        match rhs:
            case self.endtype():
                return self.__class__(fadd(self.inf, rhs), cadd(self.sup, rhs))

            case self.__class__():
                return self.__class__(fadd(self.inf, rhs.inf), cadd(self.sup, rhs.sup))

            case int():
                return self.__add__(self.__class__(rhs))

            case float():
                return self.__add__(self.converter.fromfloat(rhs))

        return NotImplemented

    def __sub__(self, rhs: Self | T | float | int) -> Self:
        csub = self.operator.csub
        fsub = self.operator.fsub

        match rhs:
            case self.endtype():
                return self.__class__(fsub(self.inf, rhs), csub(self.sup, rhs))

            case self.__class__():
                return self.__class__(fsub(self.inf, rhs.sup), csub(self.sup, rhs.inf))

            case int():
                return self.__sub__(self.__class__(rhs))

            case float():
                return self.__sub__(self.converter.fromfloat(rhs))

        return NotImplemented

    def __mul__(self, rhs: Self | T | float | int) -> Self:
        ZERO = self.operator.ZERO
        cmul = self.operator.cmul
        fmul = self.operator.fmul

        match rhs:
            case self.endtype():
                inf = min(fmul(self.inf, rhs), fmul(self.sup, rhs))
                sup = max(cmul(self.inf, rhs), cmul(self.sup, rhs))
                return self.__class__(inf, sup)

            case self.__class__():
                if self.sup <= ZERO:
                    if rhs.sup <= ZERO:
                        inf = fmul(self.sup, rhs.sup)
                        sup = cmul(self.inf, rhs.inf)
                    elif rhs.inf >= ZERO:
                        inf = fmul(self.inf, rhs.sup)
                        sup = cmul(self.sup, rhs.inf)
                    else:
                        inf = fmul(self.inf, rhs.sup)
                        sup = cmul(self.inf, rhs.inf)

                    return self.__class__(inf, sup)

                if self.inf >= ZERO:
                    if rhs.sup <= ZERO:
                        inf = fmul(self.sup, rhs.inf)
                        sup = cmul(self.inf, rhs.sup)
                    elif rhs.inf >= ZERO:
                        inf = fmul(self.inf, rhs.inf)
                        sup = cmul(self.sup, rhs.sup)
                    else:
                        inf = fmul(self.sup, rhs.inf)
                        sup = cmul(self.sup, rhs.sup)

                    return self.__class__(inf, sup)

                if rhs.sup <= ZERO:
                    inf = fmul(self.sup, rhs.inf)
                    sup = cmul(self.inf, rhs.inf)
                elif rhs.inf >= ZERO:
                    inf = fmul(self.inf, rhs.sup)
                    sup = cmul(self.sup, rhs.sup)
                else:
                    inf = min(fmul(self.inf, rhs.sup), fmul(self.sup, rhs.inf))
                    sup = max(cmul(self.inf, rhs.inf), cmul(self.sup, rhs.sup))

                return self.__class__(inf, sup)

            case int():
                return self.__mul__(self.__class__(rhs))

            case float():
                return self.__mul__(self.converter.fromfloat(rhs))

        return NotImplemented

    def __truediv__(self, rhs: Self | T | float | int) -> Self:
        ZERO = self.operator.ZERO
        cdiv = self.operator.cdiv
        fdiv = self.operator.fdiv

        match rhs:
            case self.endtype():
                if rhs == ZERO:
                    raise ZeroDivisionError

                inf = min(fdiv(self.inf, rhs), fdiv(self.sup, rhs))
                sup = max(cdiv(self.inf, rhs), cdiv(self.sup, rhs))
                return self.__class__(inf, sup)

            case self.__class__():
                if rhs.inf == rhs.sup == ZERO:
                    raise ZeroDivisionError

                if rhs.inf <= ZERO <= rhs.sup:
                    INFINITY = self.operator.INFINITY

                    if self.inf == self.sup == ZERO:
                        return self.__class__()

                    if rhs.inf != ZERO and rhs.sup != ZERO:
                        return self.__class__(-INFINITY, INFINITY)

                    if self.sup < ZERO:
                        if rhs.sup == ZERO:
                            return self.__class__(fdiv(self.sup, rhs.inf), INFINITY)

                        return self.__class__(-INFINITY, cdiv(self.sup, rhs.sup))

                    if self.sup == ZERO:
                        if rhs.sup == ZERO:
                            return self.__class__(ZERO, INFINITY)

                        return self.__class__(-INFINITY, ZERO)

                    if self.inf < ZERO:
                        return self.__class__(-INFINITY, INFINITY)

                    if self.inf == ZERO:
                        if rhs.sup == ZERO:
                            return self.__class__(-INFINITY, ZERO)

                        return self.__class__(ZERO, INFINITY)

                    if rhs.sup == ZERO:
                        return self.__class__(-INFINITY, cdiv(self.inf, rhs.sup))

                    return self.__class__(fdiv(self.inf, rhs.inf), INFINITY)

                if rhs.sup < ZERO:
                    if self.sup < ZERO:
                        inf = fdiv(self.sup, rhs.inf)
                        sup = cdiv(self.inf, rhs.sup)
                    elif self.inf > ZERO:
                        inf = fdiv(self.sup, rhs.sup)
                        sup = cdiv(self.inf, rhs.inf)
                    else:
                        inf = fdiv(self.sup, rhs.sup)
                        sup = cdiv(self.inf, rhs.sup)

                    return self.__class__(inf, sup)

                if self.sup < ZERO:
                    inf = fdiv(self.inf, rhs.inf)
                    sup = cdiv(self.sup, rhs.sup)
                elif self.inf > ZERO:
                    inf = fdiv(self.inf, rhs.sup)
                    sup = cdiv(self.sup, rhs.inf)
                else:
                    inf = fdiv(self.inf, rhs.inf)
                    sup = cdiv(self.sup, rhs.inf)

                return self.__class__(inf, sup)

            case int():
                return self.__truediv__(self.__class__(rhs))

            case float():
                return self.__truediv__(self.converter.fromfloat(rhs))

        return NotImplemented

    def __pow__(self, rhs: int) -> Self:
        ZERO = self.operator.ZERO
        ONE = self.operator.ONE

        if not isinstance(rhs, int):
            return NotImplemented

        if rhs < 0:
            return self.__pow__(-rhs).__rtruediv__(ONE)

        result = self.__class__(ONE)
        tmp = self.copy()
        is_even = rhs % 2 == 0

        while rhs != 0:
            if rhs % 2 != 0:
                result *= tmp

            rhs //= 2
            tmp *= tmp

        if is_even and self.inf <= ZERO <= self.sup:
            result.inf = ZERO

        return result

    def __and__(self, rhs: Self | T | float | int) -> Self:
        match rhs:
            case self.endtype():
                if not self.inf <= rhs <= self.sup:
                    raise ValueError

                return self.__class__(rhs)

            case self.__class__():
                inf = max(self.inf, rhs.inf)
                sup = min(self.sup, rhs.sup)

                if inf > sup:
                    raise ValueError

                return self.__class__(inf, sup)

            case int():
                return self.__and__(self.__class__(rhs))

            case float():
                return self.__and__(self.converter.fromfloat(rhs))

        return NotImplemented

    def __or__(self, rhs: Self | T | float | int) -> Self:
        match rhs:
            case self.endtype():
                return self.__class__(min(self.inf, rhs), max(self.sup, rhs))

            case self.__class__():
                return self.__class__(min(self.inf, rhs.inf), max(self.sup, rhs.sup))

            case int():
                return self.__or__(self.__class__(rhs))

            case float():
                return self.__or__(self.converter.fromfloat(rhs))

        return NotImplemented

    def __radd__(self, lhs: Self | T | float | int) -> Self:
        return self.__add__(lhs)

    def __rsub__(self, lhs: Self | T | float | int) -> Self:
        return self.__neg__().__add__(lhs)

    def __rmul__(self, lhs: Self | T | float | int) -> Self:
        return self.__mul__(lhs)

    def __rtruediv__(self, lhs: Self | T | float | int) -> Self:
        match lhs:
            case self.endtype() | float() | int():
                return self.__class__(lhs).__truediv__(self)

            case self.__class__():
                return lhs.__truediv__(self)

        return NotImplemented

    def __rand__(self, lhs: Self | T | float | int) -> Self:
        return self.__and__(lhs)

    def __ror__(self, lhs: Self | T | float | int) -> Self:
        return self.__or__(lhs)

    def __neg__(self) -> Self:
        return self.__class__(-self.sup, -self.inf)

    def __pos__(self) -> Self:
        return self.__class__(self.inf, self.sup)

    def __abs__(self) -> Self:
        ZERO = self.operator.ZERO

        if self.inf > ZERO:
            return self.__class__(self.inf, self.sup)

        if self.sup < ZERO:
            return self.__class__(-self.sup, -self.inf)

        return self.__class__(ZERO, max(-self.inf, self.sup))

    def __copy__(self) -> Self:
        return self.copy()

    def __sqrt(self):
        if self.inf < self.operator.ZERO:
            raise ValueError("math domain error")

        inf = self.operator.fsqr(self.inf)
        sup = self.operator.csqr(self.sup)
        return self.__class__(inf, sup)



"""
##############################################
Mathematical functions (:mod:`verry.function`)
##############################################

.. currentmodule:: verry.function

This module provides mathematical functions.

Constant functions
==================

.. autosummary::
    :toctree: generated/

    e
    ln2
    pi

Power, exponents, and logarithmic functions
===========================================

.. autosummary::
    :toctree: generated/

    exp
    log
    pow
    sqrt

Trigonometric functions
=======================

.. autosummary::
    :toctree: generated/

    cos
    sin
    tan

"""

import math
from typing import Any, overload

import mpmath
import mpmath.ctx_mp_python

from verry.affine import AffineForm
from verry.autodiff.autodiff import _defderiv, _primitive
from verry.interval.interval import Interval
from verry.intervalseries import IntervalSeries

_cos: Any = None
_e: Any = None
_exp: Any = None
_ln2: Any = None
_log: Any = None
_pi: Any = None
_pow: Any = None
_sin: Any = None
_sqrt: Any = None
_tan: Any = None


@overload
def cos[T: Interval](x: T, /) -> T: ...


@overload
def cos(x: float | int, /) -> float: ...


@overload
def cos(x: Any, /) -> Any: ...


@_primitive
def cos(x, /):
    """Cosine."""
    if (fun := getattr(type(x), "_verry_overload_", None)) is not None:
        if (res := fun(x, _cos, x)) is not NotImplemented:
            return res

        raise TypeError

    match x:
        case mpmath.ctx_mp_python.mpnumeric():
            return mpmath.cos(x)

        case float() | int():
            return math.cos(x)

        case _:
            raise TypeError


@overload
def e[T: Interval | AffineForm | IntervalSeries](x: T, /) -> T: ...


@overload
def e(x: float | int, /) -> float: ...


@overload
def e(x: Any, /) -> Any: ...


@_primitive
def e(x, /):
    """Napier's constant.

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> print(format(e(1.0), ".6f"))
    2.718282
    >>> print(e(FI()))
    [2.718281, 2.718282]
    """
    if fun := getattr(type(x), "_verry_overload_", None):
        if (res := fun(x, _e, x)) is not NotImplemented:
            return res

        raise TypeError

    match x:
        case mpmath.ctx_mp_python.mpnumeric():
            return mpmath.e

        case float() | int():
            return math.e

        case _:
            raise TypeError


@overload
def exp[T: Interval](x: T, /) -> T: ...


@overload
def exp(x: float | int, /) -> float: ...


@overload
def exp(x: Any, /) -> Any: ...


@_primitive
def exp(x, /):
    """Exponential.

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> print(format(exp(2), ".6f"))
    7.389056
    >>> print(exp(FI(2)))
    [7.389056, 7.389057]
    """
    if fun := getattr(type(x), "_verry_overload_", None):
        if (res := fun(x, _exp, x)) is not NotImplemented:
            return res

        raise TypeError

    match x:
        case mpmath.ctx_mp_python.mpnumeric():
            return mpmath.exp(x)

        case float() | int():
            return math.exp(x)

        case _:
            raise TypeError


@overload
def ln2[T: Interval | AffineForm | IntervalSeries](x: T, /) -> T: ...


@overload
def ln2(x: float | int, /) -> float: ...


@overload
def ln2(x: Any, /) -> Any: ...


@_primitive
def ln2(x, /):
    """Natural logarithm of 2.

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> print(format(ln2(1.0), ".6f"))
    0.693147
    >>> print(ln2(FI()))
    [0.693147, 0.693148]
    """
    if fun := getattr(type(x), "_verry_overload_", None):
        if (res := fun(x, _ln2, x)) is not NotImplemented:
            return res

        raise TypeError

    match x:
        case mpmath.ctx_mp_python.mpnumeric():
            return mpmath.ln2

        case float() | int():
            return float.fromhex("0x1.62e42fefa39efp-1")

        case _:
            raise TypeError


@overload
def log[T: Interval](x: T, /) -> T: ...


@overload
def log(x: float | int, /) -> float: ...


@overload
def log(x: Any, /) -> Any: ...


@_primitive
def log(x, /):
    """Natural logarithm.

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> print(format(log(5), ".6f"))
    1.609438
    >>> print(log(FI(5)))
    [1.609437, 1.609438]
    """
    if fun := getattr(type(x), "_verry_overload_", None):
        if (res := fun(x, _log, x)) is not NotImplemented:
            return res

        raise TypeError

    match x:
        case mpmath.ctx_mp_python.mpnumeric():
            return mpmath.log(x)

        case float() | int():
            return math.log(x)

        case _:
            raise TypeError


@overload
def pi[T: Interval | AffineForm | IntervalSeries](x: T, /) -> T: ...


@overload
def pi(x: float | int, /) -> float: ...


@overload
def pi(x: Any, /) -> Any: ...


@_primitive
def pi(x, /):
    """Pi.

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> print(format(pi(1.0), ".6f"))
    3.141593
    >>> print(pi(FI()))
    [3.141592, 3.141593]
    """
    if fun := getattr(type(x), "_verry_overload_", None):
        if (res := fun(x, _pi, x)) is not NotImplemented:
            return res

        raise TypeError

    match x:
        case mpmath.ctx_mp_python.mpnumeric():
            return mpmath.pi

        case float() | int():
            return math.pi

        case _:
            raise TypeError


@overload
def pow[T: Interval](x: T | float | int, y: T, /) -> T: ...


@overload
def pow[T: Interval](x: T, y: float | int, /) -> T: ...


@overload
def pow(x: float | int, y: float | int, /) -> float: ...


@overload
def pow(x: Any, y: Any, /) -> Any: ...


@_primitive
def pow(x, y, /):
    """`x` raised to the power `y`.

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> print(format(pow(3.25, 1.25), ".6f"))
    4.363693
    >>> print(pow(FI("3.25"), FI("1.25")))
    [4.363693, 4.363694]
    """
    linearized = (x, y)

    if type(x) is not type(y) and issubclass(type(y), type(x)):
        linearized = (y, x)

    for z in linearized:
        if fun := getattr(type(z), "_verry_overload_", None):
            if (res := fun(z, _pow, x, y)) is not NotImplemented:
                return res

    mpnumeric = mpmath.ctx_mp_python.mpnumeric

    match x, y:
        case (mpnumeric(), _) | (_, mpnumeric()):
            return mpmath.power(x, y)

        case (float() | int(), float() | int()):
            return math.pow(x, y)

        case _:
            raise TypeError


@overload
def sin[T: Interval](x: T, /) -> T: ...


@overload
def sin(x: float | int, /) -> float: ...


@overload
def sin(x: Any, /) -> Any: ...


@_primitive
def sin(x, /):
    """Sine."""
    if fun := getattr(type(x), "_verry_overload_", None):
        if (res := fun(x, _sin, x)) is not NotImplemented:
            return res

        raise TypeError

    match x:
        case mpmath.ctx_mp_python.mpnumeric():
            return mpmath.sin(x)

        case float() | int():
            return math.sin(x)

        case _:
            raise TypeError


@overload
def sqrt[T: Interval](x: T, /) -> T: ...


@overload
def sqrt(x: float | int, /) -> float: ...


@overload
def sqrt(x: Any, /) -> Any: ...


@_primitive
def sqrt(x, /):
    """Square root.

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> print(format(sqrt(2.0), ".6f"))
    1.414214
    >>> print(sqrt(FI(2)))
    [1.414213, 1.414214]
    """
    if fun := getattr(type(x), "_verry_overload_", None):
        if (res := fun(x, _sqrt, x)) is not NotImplemented:
            return res

        raise TypeError

    match x:
        case mpmath.ctx_mp_python.mpnumeric():
            return mpmath.sqrt(x)

        case float() | int():
            return math.sqrt(x)

        case _:
            raise TypeError


@overload
def tan[T: Interval](x: T, /) -> T: ...


@overload
def tan(x: float | int, /) -> float: ...


@overload
def tan(x: Any, /) -> Any: ...


@_primitive
def tan(x, /):
    """Tangent."""
    if fun := getattr(type(x), "_verry_overload_", None):
        if (res := fun(x, _tan, x)) is not NotImplemented:
            return res

        raise TypeError

    match x:
        case mpmath.ctx_mp_python.mpnumeric():
            return mpmath.tan(x)

        case float() | int():
            return math.tan(x)

        case _:
            raise TypeError


_cos = cos
_e = e
_exp = exp
_ln2 = ln2
_log = log
_pi = pi
_pow = pow
_sin = sin
_sqrt = sqrt
_tan = tan

_defderiv(cos, lambda x: -sin(x))
_defderiv(e, lambda x: x * 0)
_defderiv(exp, exp)
_defderiv(ln2, lambda x: x * 0)
_defderiv(log, lambda x: 1 / x)
_defderiv(pi, lambda x: x * 0)
_defderiv(pow, lambda x, y: y * pow(x, y) / x, argnum=0)
_defderiv(pow, lambda x, y: log(x) * pow(x, y), argnum=1)
_defderiv(sin, cos)
_defderiv(sqrt, lambda x: 1 / (2 * sqrt(x)))
_defderiv(tan, lambda x: 1 / cos(x) ** 2)



import os
import sys

project = "Verry"
copyright = "2025, R. Iwanami"
author = "R. Iwanami"

extensions = [
    "sphinx.ext.githubpages",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.extlinks",
    "sphinx_design",
    "sphinx_favicon",
]

templates_path = ["_templates"]
exclude_patterns = []

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_logo = "_static/logo.svg"
html_theme_options = {
    "logo": {"text": "Verry"},
    "navigation_depth": 1,
    "secondary_sidebar_items": ["page-toc", "sourcelink"],
}

favicons = [
    "favicon.ico",
    "logo.svg",
    {"rel": "apple-touch-icon", "href": "apple-touch-icon.png"},
]

autosummary_generate = True
autodoc_typehints = "none"

napoleon_preprocess_types = False
napoleon_attr_annotations = False
napoleon_use_ivar = True

mathjax3_config = {
    "loader": {"load": ["[tex]/mathtools"]},
    "tex": {"packages": {"[+]": ["mathtools"]}},
}

extlinks = {"doi": ("https://dx.doi.org/%s", "doi:%s")}

sys.path.insert(0, os.path.abspath("../"))



from verry.affine import AffineForm
from verry.interval import FloatInterval


def test():
    def rational(x):
        return x**2 / (x**3 + 2 * x + 1)

    var = 4 + 1e-4 * FloatInterval(-1, 1)
    assert rational(AffineForm(var)).range().issubset(rational(var))



import pytest

from verry import function as vrf
from verry.autodiff import autodiff


def test_deriv():
    deriv1 = autodiff.deriv(lambda x: (x + vrf.sin(x**2)) / x)
    deriv2 = autodiff.deriv(deriv1)
    assert pytest.approx(deriv1(1.4), 1e-5) == -1.23095
    assert pytest.approx(deriv2(1.4), 1e-5) == -3.96476


def test_grad():
    grad = autodiff.grad(vrf.pow)
    assert pytest.approx(grad(4.5, -2.2), 1e-5) == (-0.0178707, 0.0549797)

    grad = autodiff.grad(lambda x, y: vrf.exp(y / x) + 2)
    assert pytest.approx(grad(1.2, 3.5), 1e-5) == (-44.9157, 15.3997)


def test_jacobian():
    jacobian = autodiff.jacobian(lambda x, y: (vrf.sin(x * y), x**2 - vrf.cos(y)))
    matrix = jacobian(2, 3)
    assert pytest.approx(matrix[0], 1e-5) == (2.88051, 1.92034)
    assert pytest.approx(matrix[1], 1e-5) == (4.00000, 0.14112)



"""
############################
Typing (:mod:`verry.typing`)
############################

This module provides type definitions commonly used between modules.

.. autoclass:: Scalar
    :show-inheritance:
    :no-members:

.. autoclass:: SignedComparable
    :show-inheritance:
    :no-members:

.. autoclass:: ComparableScalar
    :show-inheritance:
    :no-members:

"""

from abc import abstractmethod
from typing import Protocol, Self, SupportsAbs


class Scalar(Protocol):
    """Protocol that ensures scalar-like behavior.

    Objects implementing this protocol must have four arithmetic operations and
    integer power defined, and four arithmetic operations must be compatible with
    integers.
    """

    __slots__ = ()

    @abstractmethod
    def __add__(self, rhs: Self | int) -> Self: ...

    @abstractmethod
    def __sub__(self, rhs: Self | int) -> Self: ...

    @abstractmethod
    def __mul__(self, rhs: Self | int) -> Self: ...

    @abstractmethod
    def __truediv__(self, rhs: Self | int) -> Self: ...

    @abstractmethod
    def __pow__(self, rhs: int) -> Self: ...

    @abstractmethod
    def __radd__(self, lhs: Self | int) -> Self: ...

    @abstractmethod
    def __rsub__(self, lhs: Self | int) -> Self: ...

    @abstractmethod
    def __rmul__(self, lhs: Self | int) -> Self: ...

    @abstractmethod
    def __rtruediv__(self, lhs: Self | int) -> Self: ...

    @abstractmethod
    def __neg__(self) -> Self: ...

    @abstractmethod
    def __pos__(self) -> Self: ...


class SignedComparable(SupportsAbs, Protocol):
    """Protocol that ensures signed comparability."""

    __slots__ = ()

    @abstractmethod
    def __lt__(self, rhs: Self) -> bool: ...

    @abstractmethod
    def __le__(self, rhs: Self) -> bool: ...

    @abstractmethod
    def __gt__(self, rhs: Self) -> bool: ...

    @abstractmethod
    def __ge__(self, rhs: Self) -> bool: ...

    @abstractmethod
    def __neg__(self) -> Self: ...

    @abstractmethod
    def __pos__(self) -> Self: ...


class ComparableScalar(SignedComparable, Scalar, Protocol):
    """Intersection of :class:`SignedComparable` and :class:`Scalar`."""

    __slots__ = ()



from .function import cos, exp, log, pow, sin, tan
from .interval import FloatInterval, Interval

__all__ = [
    "cos",
    "exp",
    "log",
    "pow",
    "sin",
    "tan",
    "Interval",
    "FloatInterval",
]



from collections.abc import Callable, Sequence

from verry.autodiff.autodiff import jacobian
from verry.interval.interval import Interval
from verry.intervalseries import IntervalSeries, localcontext
from verry.linalg.intervalmatrix import IntervalMatrix


def seriessolution[T: Interval](
    fun: Callable, t0: T, y0: IntervalMatrix[T] | Sequence[T], deg: int
) -> tuple[IntervalSeries[T], ...]:
    """Return the Taylor polynomial of the ODE solution.

    Parameters
    ----------
    fun : Callable
        Right-hand side of the system.
    t0 : Interval
        Initial time.
    y0 : IntervalMatrix | Sequence[Interval]
        Initial state.
    deg : int
        Degree of Taylor polynomial.

    Returns
    -------
    tuple[IntervalSeries, ...]

    Warnings
    --------
    `fun` must neither be a constant nor contain conditional branches
    (cf. :doc:`/userguide/pitfall`).

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> series = seriessolution(lambda t, y: (y,), FI(0), [FI(1)], 4)
    >>> print([format(coeff.mid(), ".3f") for coeff in series[0].coeffs])
    ['1.000', '1.000', '0.500', '0.167', '0.042']
    """
    t = IntervalSeries([t0, 1], intvl=type(t0))
    y = (IntervalSeries([x], intvl=type(t0)) for x in y0)

    with localcontext(rounding="TYPE1", deg=0) as ctx:
        while ctx.deg < deg:
            y = (y0 + dy.integrate() for y0, dy in zip(y0, fun(t, *y)))
            ctx.deg += 1

    return tuple(y)


def variationaleq(fun: Callable, sol: Callable) -> Callable:
    r"""Return the right-hand side of variational equations.

    Parameters
    ----------
    fun : Callable
        Right-hand side of the system.
    sol : Callable
        Solution of ODEs.

    Returns
    -------
    Callable

    Warnings
    --------
    `fun` must neither be a constant nor contain conditional branches
    (cf. :doc:`/userguide/pitfall`).

    Notes
    -----
    Variational equations of ODEs :math:`\mathrm{d}x/\mathrm{d}t=f(t,x)` are defined as

    .. math:: \frac{\mathrm{d}v}{\mathrm{d}t} = (v\cdot\nabla)f(t,x),

    where :math:`x=\phi(t)` is a solution of ODEs.
    """

    def result(t, *v):
        n = len(v)
        jac = jacobian(lambda *y: fun(t, *y))(*sol(t))
        return tuple(sum(jac[i][j] * v[j] for j in range(n)) for i in range(n))

    return result



def cadd(lhs: float, rhs: float) -> float: ...
def cmul(lhs: float, rhs: float) -> float: ...
def cdiv(lhs: float, rhs: float) -> float: ...
def csqr(value: float) -> float: ...
def fsqr(value: float) -> float: ...



import dataclasses
import inspect
import itertools
from collections.abc import Callable, Sequence
from typing import Any, Literal, TypeIs, overload

from verry.integrate.integrator import Integrator, kashi
from verry.integrate.tracker import Tracker, doubletontracker
from verry.integrate.utility import seriessolution, variationaleq
from verry.interval.interval import Interval
from verry.intervalseries import IntervalSeries
from verry.linalg.intervalmatrix import IntervalMatrix


class AbortSolving(Exception):
    """Raised by a callback function to abort solvers.

    Parameters
    ----------
    message : str, default="aborted"
    """

    message: str

    def __init__(self, message="aborted", *args, **kwargs):
        super().__init__(message, *args, **kwargs)
        self.message = message


class OdeSolution[T: Interval]:
    """Solution of ODEs.

    Attributes
    ----------
    ts : list, length n+1
        Segment breakpoints.
    series : list[tuple[IntervalSeries, ...]], length n
        Sequence of solutions for each segment.
    """

    __slots__ = ("ts", "series")
    ts: list
    series: list[tuple[IntervalSeries[T], ...]]

    def __init__(self, ts: Sequence, series: Sequence[tuple[IntervalSeries[T], ...]]):
        self.ts = list(ts)
        self.series = list(series)

    def __call__(self, arg: Any) -> tuple[T, ...]:
        """Return the value that the solution takes in `arg`."""
        if isinstance(arg, str):
            raise TypeError

        intvl = self.series[0][0].interval
        arg = intvl.ensure(arg)
        result: list[T] | None = None

        for ts, series in zip(itertools.pairwise(self.ts), self.series):
            if (domain := intvl(ts[0], ts[1])).isdisjoint(arg):
                continue

            if result is None:
                result = [x.eval((arg & domain) - ts[0]) for x in series]
            else:
                for i, x in enumerate(series):
                    result[i] |= x.eval((arg & domain) - ts[0])

        if result is None:
            raise ValueError

        return tuple(result)


@dataclasses.dataclass(frozen=True, slots=True)
class SolverResult[T1, T2: Literal["ABORTED", "FAILURE", "SUCCESS"]]:
    """Output of ODE solvers.

    Attributes
    ----------
    status : Literal["ABORTED", "FAILURE", "SUCCESS"]
    content
    message : str
        Report from the solver. Typically a reason for a failure.
    """

    status: T2
    content: T1
    message: str


@dataclasses.dataclass(frozen=True, slots=True)
class C0SolverResultContent[T: Interval]:
    """Information obtained when :class:`C0Solver` successfully worked.

    Attributes
    ----------
    t : Interval
        Boundary time passed to :meth:`C0Solver.solve`.
    y : tuple[Interval, ...], length n
        The value of ``sol(t)``.
    sol : OdeSolution
        Solution of ODEs.
    """

    t: T
    y: tuple[T, ...]
    sol: OdeSolution[T]


@dataclasses.dataclass(frozen=True, slots=True)
class C0SolverCallbackArg[T: Interval]:
    """Argument of callback functions passed to :meth:`C0Solver.solve`.

    Attributes
    ----------
    t : Interval
        Current time.
    y : tuple[Interval, ...]
        Current state.
    t_prev : Interval
        Previous time.
    series : tuple[IntervalSeries, ...]
        Current interval series.
    """

    t: T
    y: tuple[T, ...]
    t_prev: T
    series: tuple[IntervalSeries[T], ...]


class C0Solver:
    """ODE solver.

    Parameters
    ----------
    integrator : Callable[[], type[Integrator]] | type[Integrator], optional
        The default is :func:`kashi`.
    tracker : type[Tracker], optional
        The default is :func:`doubletontracker`.

    Notes
    -----
    The choice of `tracker` has a significant impact on accuracy and computation time.
    Our recommendations are as follows:

    1. Use :func:`doubletontracker` at first.
    2. Use :func:`affinetracker` if the solver cannot compute solutions, or the accuracy
       of solutions is not sufficient.
    """

    __slots__ = ("_integrator", "_tracker")
    _integrator: type[Integrator]
    _tracker: type[Tracker]

    def __init__(
        self,
        integrator: Callable[[], type[Integrator]] | type[Integrator] | None = None,
        tracker: Callable[[], type[Tracker]] | type[Tracker] | None = None,
    ):
        if integrator is None:
            integrator = kashi()
        elif not _is_integrator(integrator):
            integrator = integrator()

        if tracker is None:
            tracker = doubletontracker()
        elif not _is_tracker(tracker):
            tracker = tracker()

        if not issubclass(integrator, Integrator):
            raise TypeError

        if not issubclass(tracker, Tracker):
            raise TypeError

        self._integrator = integrator
        self._tracker = tracker

    def solve[T: Interval](
        self,
        fun: Callable,
        t0: T,
        y0: IntervalMatrix[T] | Sequence[T],
        t_bound: T,
        callback: Callable[[C0SolverCallbackArg[T]], None] | None = None,
    ) -> (
        SolverResult[None, Literal["ABORTED"]]
        | SolverResult[None, Literal["FAILURE"]]
        | SolverResult[C0SolverResultContent[T], Literal["SUCCESS"]]
    ):
        r"""Solve an initial value problem for a system of ODEs.

        Parameters
        ----------
        fun : Callable
            Right-hand side of the system. The calling signature is ``fun(t, *y)``.
        t0 : Interval
            Initial time.
        y0 : IntervalMatrix | Sequence[Interval]
            Initial state.
        t_bound : Interval
            Boundary time. Its infimum must be greater than the supremum of `t0`.
        callback : Callable[[C0SolverCallbackArg], None], optional
            Callback function called at each step. You can abort the solver by raising
            :class:`AbortSolving`.

        Returns
        -------
        r : SolverResult
            If ``r.status`` is ``"SUCCESS"``, ``r.content`` is
            :class:`C0SolverResultContent`; otherwise, ``r.content`` is ``None``.

        Warnings
        --------
        `fun` must be a :math:`C^\infty`-function on some open interval that
        contains `t0` and `t_bound`. Furthermore, `fun` must neither be a constant nor
        contain conditional branches (cf. :doc:`/userguide/pitfall`).

        Examples
        --------
        This example computes :math:`\cos(x)` and :math:`\sin(x)`.

        >>> from verry import FloatInterval as FI
        >>> from verry import function as vrf
        >>> PI = vrf.pi(FI())
        >>> solver = C0Solver()
        >>> r = solver.solve(lambda t, x, y: (y, -x), FI(0), [FI(1), FI(0)], PI)
        >>> print(r.status)
        SUCCESS
        >>> print(r.content.y[0])
        [-1.000001, -0.999999]

        The next example fails due to the blow-up of the solution.

        >>> from verry.integrate import eilo, affinetracker
        >>> solver = C0Solver(eilo(min_step=1e-4), affinetracker)
        >>> r = solver.solve(lambda t, x: (x**2,), FI(0), [FI(1)], FI(2))
        >>> print(r.status)
        FAILURE
        >>> print(r.message)
        failed to determine a step size
        """
        if not isinstance(t0, Interval):
            raise TypeError

        if not isinstance(y0, IntervalMatrix):
            y0 = IntervalMatrix(y0, intvl=type(t0))

        intvl = type(t0)
        intvlmat = type(y0)
        eye = intvlmat.eye(len(y0), intvl=intvl)

        ts = [t0.inf]
        series = []

        tracker = self._tracker(y0)
        itor = self._integrator(fun, t0, tracker.hull(), t_bound)

        while itor.status == "RUNNING":
            if not (res := itor.step())[0]:
                return SolverResult("FAILURE", None, res[1])

            u0 = itor.t_prev
            u1 = itor.t

            varfun = variationaleq(fun, lambda t: tuple(x(t - u0) for x in itor.series))  # type: ignore
            tmp = seriessolution(fun, u0, tracker.sample(), itor.order - 1)
            a1 = intvlmat([x.eval(u1 - u0) for x in tmp], intvl=intvl)
            jac = eye.empty_like()

            for i in range(len(y0)):
                tmp = seriessolution(varfun, u0, eye[i], itor.order - 1)
                a1[i] += itor.series[i].coeffs[-1] * (u1 - u0) ** itor.order  # type: ignore
                jac[:, i] = intvlmat([x.eval(u1 - u0) for x in tmp], intvl=intvl)

            tracker.update(jac, a1)
            itor.update(u1, y1 := tracker.hull())

            ts.append(u1.sup)
            series.append(itor.series)

            if callback is not None:
                try:
                    callback(C0SolverCallbackArg(u1, y1, u0, itor.series))  # type: ignore
                except AbortSolving as exc:
                    return SolverResult("ABORTED", None, exc.message)

        sol = OdeSolution(ts, series)  # type: ignore
        content = C0SolverResultContent(itor.t, itor.y, sol)
        return SolverResult("SUCCESS", content, "success")


@dataclasses.dataclass(frozen=True, slots=True)
class C1SolverResultContent[T1: Interval, T2: IntervalMatrix[T1] = IntervalMatrix[T1]]:  # type: ignore
    # TODO: This "type: ignore" must be removed when mypy supports PEP 696.
    #       See https://github.com/python/mypy/issues/14851
    """Information obtained when :class:`C1Solver` successfully worked.

    Attributes
    ----------
    t : Interval
        Boundary time passed to :meth:`C1Solver.solve`.
    y : tuple[Interval, ...], length n
        The value of ``sol(t)``.
    sol : OdeSolution
        Solution of ODEs.
    jac : IntervalMatrix, shape (n, n)
        Jacobian matrix of flow with respect to initial values.
    """

    t: T1
    y: tuple[T1, ...]
    sol: OdeSolution[T1]
    jac: T2


@dataclasses.dataclass(frozen=True, slots=True)
class C1SolverCallbackArg[T1: Interval, T2: IntervalMatrix[T1] = IntervalMatrix[T1]]:  # type: ignore
    # TODO: This "type: ignore" must be removed when mypy supports PEP 696.
    #       See https://github.com/python/mypy/issues/14851
    """Argument of callback functions passed to :meth:`C1Solver.solve`.

    Attributes
    ----------
    t : Interval
        Current time.
    y : tuple[Interval, ...]
        Current state.
    t_prev : Interval
        Previous time.
    series : tuple[IntervalSeries, ...]
        Current interval series.
    jac : IntervalMatrix
        Jacobian matrix of flow in the current step.
    totjac : IntervalMatrix
        Jacobian matrix of flow with respect to initial values.
    """

    t: T1
    y: tuple[T1, ...]
    t_prev: T1
    series: tuple[IntervalSeries[T1], ...]
    jac: T2
    totjac: T2


class C1Solver:
    """ODE solver using variational equations.

    Parameters
    ----------
    integrator : Callable[[], type[Integrator]] | type[Integrator], optional
        The default is :func:`kashi`.
    tracker : type[Tracker], optional
        The default is :func:`doubletontracker`.

    Notes
    -----
    The choice of `tracker` has a significant impact on accuracy and computation time.
    Our recommendations are as follows:

    1. Use :func:`doubletontracker` at first.
    2. Use :func:`affinetracker` if the solver cannot compute solutions, or the
       accuracy of solutions is not sufficient.
    """

    __slots__ = ("_integrator", "_tracker")
    _integrator: type[Integrator]
    _tracker: type[Tracker]

    def __init__(
        self,
        integrator: Callable[[], type[Integrator]] | type[Integrator] | None = None,
        tracker: Callable[[], type[Tracker]] | type[Tracker] | None = None,
    ):
        if integrator is None:
            integrator = kashi()
        elif not _is_integrator(integrator):
            integrator = integrator()

        if tracker is None:
            tracker = doubletontracker()
        elif not _is_tracker(tracker):
            tracker = tracker()

        if not issubclass(integrator, Integrator):
            raise TypeError

        if not issubclass(tracker, Tracker):
            raise TypeError

        self._integrator = integrator
        self._tracker = tracker

    @overload
    def solve[T: Interval](
        self,
        fun: Callable,
        t0: T,
        y0: Sequence[T],
        t_bound: T,
        callback: Callable[[C1SolverCallbackArg[T]], None] | None = ...,
    ) -> (
        SolverResult[None, Literal["ABORTED"]]
        | SolverResult[None, Literal["FAILURE"]]
        | SolverResult[C1SolverResultContent[T], Literal["SUCCESS"]]
    ): ...

    @overload
    def solve[T1: Interval, T2: IntervalMatrix[T1]](  # type: ignore
        # TODO: This "type: ignore" must be removed when mypy supports PEP 696.
        #       See https://github.com/python/mypy/issues/14851
        self,
        fun: Callable,
        t0: T1,
        y0: T2,
        t_bound: T1,
        callback: Callable[[C1SolverCallbackArg[T1, T2]], None] | None = ...,
    ) -> (
        SolverResult[None, Literal["ABORTED"]]
        | SolverResult[None, Literal["FAILURE"]]
        | SolverResult[C1SolverResultContent[T1, T2], Literal["SUCCESS"]]
    ): ...

    def solve(self, fun, t0, y0, t_bound, callback=None):
        r"""Solve an initial value problem for a system of ODEs.

        Parameters
        ----------
        fun : Callable
            Right-hand side of the system. The calling signature is ``fun(t, *y)``.
        t0 : Interval
            Initial time.
        y0 : IntervalMatrix | Sequence[Interval]
            Initial state.
        t_bound : Interval
            Boundary time. Its infimum must be greater than the supremum of `t0`.
        callback : Callable[[C1SolverCallbackArg], None], optional
            Callback function called at each step. You can abort the solver by raising
            :class:`AbortSolving`.

        Returns
        -------
        r : SolverResult
            If ``r.status`` is ``"SUCCESS"``, ``r.content`` is
            :class:`C1SolverResultContent`; otherwise, ``r.content`` is ``None``.

        Warnings
        --------
        `fun` must be a :math:`C^\infty`-function on some open interval that
        contains `t0` and `t_bound`. Furthermore, `fun` must neither be a constant nor
        contain conditional branches (cf. :doc:`/userguide/pitfall`).

        Examples
        --------
        This example computes :math:`\cos(x)` and :math:`\sin(x)`.

        >>> from verry import FloatInterval as FI
        >>> from verry import function as vrf
        >>> PI = vrf.pi(FI())
        >>> solver = C1Solver()
        >>> r = solver.solve(lambda t, x, y: (y, -x), FI(0), [FI(1), FI(0)], PI)
        >>> print(r.status)
        SUCCESS
        >>> print(r.content.y[0])
        [-1.000001, -0.999999]

        The next example fails due to the blow-up of the solution.

        >>> from verry.integrate import eilo, affinetracker
        >>> solver = C1Solver(eilo(min_step=1e-4), affinetracker)
        >>> r = solver.solve(lambda t, x: (x**2,), FI(0), [FI(1)], FI(2))
        >>> print(r.status)
        FAILURE
        >>> print(r.message)
        failed to determine a step size
        """
        if not isinstance(t0, Interval):
            raise TypeError

        if not isinstance(y0, IntervalMatrix):
            y0 = IntervalMatrix(y0, intvl=type(t0))

        intvl = type(t0)
        intvlmat = type(y0)
        eye = intvlmat.eye(len(y0), intvl=intvl)

        ts = [t0.inf]
        series = []
        totjac = intvlmat.eye(len(y0), intvl=intvl)

        tracker = self._tracker(y0)
        miditor = self._integrator(fun, t0, tracker.sample(), t_bound)
        itor = self._integrator(fun, t0, tracker.hull(), t_bound)

        while miditor.status == "RUNNING" and itor.status == "RUNNING":
            if not (res := miditor.step())[0]:
                return SolverResult("FAILURE", None, res[1])

            if not (res := itor.step())[0]:
                return SolverResult("FAILURE", None, res[1])

            u0 = itor.t_prev
            u1 = t_bound

            if miditor.status == "RUNNING" or itor.status == "RUNNING":
                u1 = intvl(min(itor.t.sup, miditor.t.sup))

            varfun = variationaleq(fun, lambda t: tuple(x(t - u0) for x in itor.series))
            varitors = [self._integrator(varfun, u0, x, u1) for x in eye]

            for x in varitors:
                if not (res := x.step())[0]:
                    return SolverResult("FAILURE", None, res[1])

            if any(x.status == "RUNNING" for x in varitors):
                u1 = intvl(min(x.t.sup for x in varitors))

            a1 = intvlmat([x.eval(u1 - u0) for x in miditor.series], intvl=intvl)
            jac = totjac.empty_like()

            for i in range(len(y0)):
                tmp = [x.eval(u1 - u0) for x in varitors[i].series]
                jac[:, i] = intvlmat(tmp, intvl=intvl)

            tracker.update(jac, a1)
            miditor.update(u1, tracker.sample())
            itor.update(u1, y1 := tracker.hull())

            ts.append(u1.sup)
            series.append(itor.series)
            totjac = jac @ totjac

            if callback is not None:
                try:
                    callback(C1SolverCallbackArg(u1, y1, u0, itor.series, jac, totjac))
                except AbortSolving as exc:
                    return SolverResult("ABORTED", None, exc.message)

        sol = OdeSolution(ts, series)
        content = C1SolverResultContent(itor.t, itor.y, sol, totjac)
        return SolverResult("SUCCESS", content, "success")


def _is_integrator(x: object) -> TypeIs[type[Integrator]]:
    return inspect.isclass(x) and issubclass(x, Integrator)


def _is_tracker(x: object) -> TypeIs[type[Tracker]]:
    return inspect.isclass(x) and issubclass(x, Tracker)



import inspect
from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Any, ClassVar, TypeIs

import numpy.typing as npt

from verry.affine import AffineForm, Context, getcontext, setcontext, summarize
from verry.linalg.intervalmatrix import IntervalMatrix, approx_norm, approx_qr, inv


class Tracker[T: IntervalMatrix](ABC):
    """Abstract base class for trackers.

    Each instance of :class:`Tracker` corresponds to a pair of a point and some closed
    convex set containing that point.

    Parameters
    ----------
    x0 : IntervalMatrix, shape (n,)
        Initial values.
    """

    __slots__ = ()

    @abstractmethod
    def __init__(self, x0: T):
        raise NotImplementedError

    @abstractmethod
    def hull(self) -> T:
        """Return an interval hull of the current set.

        Returns
        -------
        IntervalMatrix, shape (n,)
        """
        raise NotImplementedError

    @abstractmethod
    def sample(self) -> T:
        """Return the current point.

        Returns
        -------
        IntervalMatrix, shape (n,)
        """
        raise NotImplementedError

    @abstractmethod
    def update(self, a: T, b: T) -> None:
        r"""Update the current pair.

        Parameters
        ----------
        a : IntervalMatrix, shape (n, n)
            The image of :meth:`hull` mapped by :math:`Df(x)`.
        b : IntervalMatrix, shape (n,)
            The value obtained by mapping :meth:`sample` with :math:`f(x)`.

        Notes
        -----
        This method updates a pair :math:`(\hat{c},C)\to(\hat{c}',C')` with the
        following rule:

        .. math::

            C' \supseteq \{A(x-\hat{c})+b\mid A\in[A],\,b\in[b],\,x\in C\}.

        Note that :math:`C'\supseteq f(C)` holds by the mean value theorem.
        """
        raise NotImplementedError


def affinetracker[T: IntervalMatrix = Any](
    n: int | None = None, m: int = 0
) -> type[Tracker[T]]:
    """Tracker using affine arithmetic, usually producing the most accurate result."""

    class Result(_affinetracker[T], n=n, m=m):
        __slots__ = ()

    return Result


class _affinetracker[T: IntervalMatrix](Tracker[T], ABC):
    __slots__ = ("_context", "_current", "_matrix")
    _cls_n: ClassVar[int | None]
    _cls_m: ClassVar[int]
    _context: Context
    _current: list[AffineForm]
    _matrix: type[T]

    def __init__(self, x0):
        ctx = getcontext()
        self._context = ctx.copy()
        self._matrix = type(x0)
        setcontext(self._context)

        try:
            self._current = [AffineForm(x) for x in x0]
        finally:
            setcontext(ctx)

    def hull(self) -> T:
        intvl = self._current[0].interval
        return self._matrix([x.range() for x in self._current], intvl=intvl)

    def sample(self) -> T:
        intvl = self._current[0].interval
        return self._matrix([x.mid() for x in self._current], intvl=intvl)

    def update(self, a: T, b: T) -> None:
        """Update the current pair.

        Parameters
        ----------
        a : IntervalMatrix, shape (n, n)
            The image of :meth:`hull` mapped by :math:`Df(x)`.
        b : IntervalMatrix, shape (n,)
            The value obtained by mapping :meth:`sample` with :math:`f(x)`.
        """
        ctx = getcontext()
        setcontext(self._context)

        try:
            curr = self._current
            next: list[AffineForm] = []
            n = len(curr)

            for i in range(n):
                curr[i] -= curr[i].mid()

            for i in range(n):
                tmp = AffineForm(b[i])
                tmp += sum(a[i, j] * curr[j] for j in range(n))
                next.append(tmp)

            if self._cls_n is not None:
                summarize(next, self._cls_n, self._cls_m)

            self._current = next
        finally:
            setcontext(ctx)

    def __init_subclass__(cls, /, n: int | None, m: int, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._cls_n = n
        cls._cls_m = m


class directtracker[T: IntervalMatrix](Tracker[T]):
    """The most obvious tracker."""

    __slots__ = ("_current",)
    _current: T

    def __init__(self, x0: T):
        self._current = x0

    def hull(self) -> T:
        return self._current.copy()

    def sample(self) -> T:
        curr = self._current
        return curr.__class__(curr.mid(), intvl=curr.interval)

    def update(self, a: T, b: T) -> None:
        """Update the current pair.

        Parameters
        ----------
        a : IntervalMatrix, shape (n, n)
            The image of :meth:`hull` mapped by :math:`Df(x)`.
        b : IntervalMatrix, shape (n,)
            The value obtained by mapping :meth:`sample` with :math:`f(x)`.
        """
        curr = self._current
        self._current = a @ (curr - curr.mid()) + b


class qrtracker[T: IntervalMatrix](Tracker[T]):
    """Tracker using a QR decomposition.

    This is an implementation of Evaluation 3 in [#Lo87]_, known as a Lohner's QR
    method.

    References
    ----------
    .. [#Lo87] R. J. Lohner, "Enclosing the Solutions of Ordinary Initial and Boundary
        Value Problem," in *Computerarithmetic*, E. Kaucher, U. Kulisch, and Ch.
        Ullrich, Eds. Stuttgart, Germany: B. G. Teubner, 1987, pp. 225--286.
    """

    __slots__ = ("_b", "_m", "_r")
    _b: T
    _m: npt.NDArray
    _r: T

    def __init__(self, x0: T):
        self._b = x0.eye(len(x0), intvl=x0.interval)
        self._m = x0.mid()
        self._r = x0 - self._m

    def hull(self) -> T:
        return self._b @ self._r + self._m

    def sample(self) -> T:
        return self._b.__class__(self._m, intvl=self._b.interval)

    def update(self, a: T, b: T) -> None:
        """Update the current pair.

        Parameters
        ----------
        a : IntervalMatrix, shape (n, n)
            The image of :meth:`hull` mapped by :math:`Df(x)`.
        b : IntervalMatrix, shape (n,)
            The value obtained by mapping :meth:`sample` with :math:`f(x)`.
        """
        q0 = self._b
        tmp = a @ q0

        n = len(b)
        s = [(i, approx_norm(tmp[:, i], "two") * self._r[i].diam()) for i in range(n)]
        s.sort(key=lambda x: x[1], reverse=True)
        inf = a.inf[:, [x[0] for x in s]]
        sup = a.sup[:, [x[0] for x in s]]
        tmp = a.__class__(inf, sup, intvl=q0.interval) @ q0

        q1 = q0.__class__(approx_qr(tmp)[0], intvl=q0.interval)
        p1 = inv(q1, q1.mid().T)
        self._b = q1
        self._m = b.mid()
        self._r = (p1 @ a @ q0) @ self._r + p1 @ (b - self._m)


def doubletontracker[T: IntervalMatrix = Any](
    tracker: Callable[[], type[Tracker]] | type[Tracker] | None = None,
) -> type[Tracker[T]]:
    """Tracker suitable for large initial intervals.

    This is an implementation of Evaluation 4 in [#Lo87]_, and the name "doubleton" is
    from [#MrZg00]_.

    Parameters
    ----------
    tracker : type[Tracker], optional
        The default is :class:`qrtracker`.

    Returns
    -------
    type[Tracker]

    References
    ----------
    .. [#Lo87] R. J. Lohner, "Enclosing the Solutions of Ordinary Initial and Boundary
        Value Problem," in *Computerarithmetic*, E. Kaucher, U. Kulisch, and Ch.
        Ullrich, Eds. Stuttgart, Germany: B. G. Teubner, 1987, pp. 225--286.
    .. [#MrZg00] M. Mrozek and P. Zgliczy\u0144ski, "Set arithmetic and the enclosing
        problem in dynamics," *Ann. Pol. Math.*, vol. 74, pp. 237--259, 2000,
        :doi:`10.4064/ap-74-1-237-259`.
    """

    def is_tracker(x: object) -> TypeIs[type[Tracker]]:
        return inspect.isclass(x) and issubclass(x, Tracker)

    if tracker is None:
        tracker = qrtracker
    elif not is_tracker(tracker):
        tracker = tracker()

    if not issubclass(tracker, Tracker):
        raise TypeError

    class Result(_doubletontracker, tracker=tracker):
        __slots__ = ()

    return Result


class _doubletontracker[T: IntervalMatrix](Tracker[T], ABC):
    __slots__ = ("_c", "_m", "_r0", "_tracker")
    _cls_tracker: ClassVar[type[Tracker]]
    _tracker: Tracker[T]
    _c: T
    _m: npt.NDArray
    _r0: T

    def __init__(self, x0: T):
        self._c = x0.eye(len(x0), intvl=x0.interval)
        self._m = x0.mid()
        self._r0 = x0 - self._m
        self._tracker = self._cls_tracker(x0.zeros_like())

    def hull(self) -> T:
        return self._m + self._tracker.hull() + self._c @ self._r0

    def sample(self):
        return self._c.__class__(self._m, intvl=self._c.interval)

    def update(self, a: T, b: T) -> None:
        """Update the current pair.

        Parameters
        ----------
        a : IntervalMatrix, shape (n, n)
            The image of :meth:`hull` mapped by :math:`Df(x)`.
        b : IntervalMatrix, shape (n,)
            The value obtained by mapping :meth:`sample` with :math:`f(x)`.
        """
        c0 = self._c
        c1 = a.mid() @ c0.inf
        self._c = c0.__class__(c1, intvl=c0.interval)
        self._m = b.mid()
        tmp = b - self._m + (a @ c0 - c1) @ self._r0 + a @ self._tracker.sample()
        self._tracker.update(a, tmp)

    def __init_subclass__(cls, /, tracker: type[Tracker], **kwargs):
        super().__init_subclass__(**kwargs)
        cls._cls_tracker = tracker



"""
##############################################
Numerical linear algebra (:mod:`verry.linalg`)
##############################################

.. currentmodule:: verry.linalg

This module provides rigorous numerical linear algebra.

Matrices
========

.. autosummary::
    :toctree: generated/

    IntervalMatrix
    FloatIntervalMatrix

Operations
==========

.. autosummary::
    :toctree: generated/

    approx_inv
    approx_norm
    approx_qr
    approx_solve
    inv
    norm
    solve

Miscellaneous
=============

.. autosummary::
    :toctree: generated/

    LinAlgError

"""

from .floatintervalmatrix import FloatIntervalMatrix
from .intervalmatrix import (
    IntervalMatrix,
    LinAlgError,
    approx_inv,
    approx_norm,
    approx_qr,
    approx_solve,
    inv,
    norm,
    solve,
)

__all__ = [
    "FloatIntervalMatrix",
    "IntervalMatrix",
    "LinAlgError",
    "approx_inv",
    "approx_norm",
    "approx_qr",
    "approx_solve",
    "inv",
    "norm",
    "solve",
]



import itertools
from typing import Callable

from verry.autodiff.autodiff import deriv
from verry.interval.interval import Interval


def cumulative_trapezoid[T: Interval](fun: Callable, a: T, b: T, n: int) -> T:
    """Cumulatively integrate `fun` using the composite trapezoidal rule.

    Parameters
    ----------
    fun : Callable
        Integrand. `fun` must be an univariate scalar-valued function.
    a : Interval
        Lower limit of integration.
    b : Interval
        Upper limit of integration.
    n : int
        Number of divisions.

    Warnings
    --------
    `fun` must be a :math:`C^2`-function on some open interval that contains `a` and
    `b`. Furthermore, `fun` must neither be a constant nor contain conditional branches
    (cf. :doc:`/userguide/pitfall`).

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> from verry import function as vrf
    >>> s = cumulative_trapezoid(lambda x: x**2 + 2 * vrf.sqrt(x), FI(1), FI(2), 300)
    >>> print(s)
    [4.771236, 4.771237]
    """
    if not (isinstance(a, Interval) and type(a) is type(b)):
        raise TypeError

    if a.sup >= b.inf:
        raise ValueError

    intvl = type(a)
    h = (b - a) / n

    approx = fun(a) + fun(b)
    approx += 2 * sum(fun(a + h * i) for i in range(1, n))
    approx *= h / 2

    mesh = (x | y for x, y in itertools.pairwise(a + h * i for i in range(n + 1)))
    dfun = deriv(deriv(fun))
    error = -intvl.hull(*(dfun(x) for x in mesh)) * (b - a) ** 3 / (12 * n**2)

    return approx + error


def cumulative_simpson[T: Interval](fun: Callable, a: T, b: T, n: int) -> T:
    """Cumulatively integrate `fun` using the composite Simpson's 1/3 rule.

    Parameters
    ----------
    fun : Callable
        Integrand. `fun` must be an univariate scalar-valued function.
    a : Interval
        Lower limit of integration.
    b : Interval
        Upper limit of integration.
    n : int
        Number of divisions.

    Warnings
    --------
    `fun` must be a :math:`C^4`-function on some open interval that contains `a` and
    `b`. Furthermore, `fun` must neither be a constant nor contain conditional branches
    (cf. :doc:`/userguide/pitfall`).

    Examples
    --------
    >>> from verry import FloatInterval as FI
    >>> from verry import function as vrf
    >>> s = cumulative_simpson(lambda x: x**2 + 2 * vrf.sqrt(x), FI(1), FI(2), 10)
    >>> print(s)
    [4.771236, 4.771237]
    """
    if not (isinstance(a, Interval) and type(a) is type(b)):
        raise TypeError

    if a.sup >= b.inf:
        raise ValueError

    intvl = type(a)
    HALF = intvl(1) / 2
    h = (b - a) / n

    approx = fun(a) + 4 * fun(a + h * (n - HALF)) + fun(b)
    approx += 2 * sum(fun(a + h * i) for i in range(1, n))
    approx += 4 * sum(fun(a + h * (i - HALF)) for i in range(1, n))
    approx *= h / 6

    mesh = (x | y for x, y in itertools.pairwise(a + h * i for i in range(n + 1)))
    dfun = deriv(deriv(deriv(deriv(fun))))
    error = -intvl.hull(*(dfun(x) for x in mesh)) * (b - a) ** 5 / (2880 * n**4)

    return approx + error
